#include <nanobind/nanobind.h>
#include <nanobind/ndarray.h>
#include <nanobind/stl/function.h>
#include <nanobind/stl/tuple.h>

namespace nb = nanobind;

extern "C" {
    typedef double(*invert_singular_values_t)(double);
    void compute_singular_values_python(int rows, int cols, double *matrix, int print_errors, double *sv, int *info);
    void compute_singular_vectors_python(int rows, int cols, double *matrix, int print_errors, double *sv, double *left, double *right, int *info);
    void solve_system_python(int rows, int cols, double *matrix, double *rhs, invert_singular_values_t invert_sv_callback, int print_errors, double *solution, int *info);
}

using vector_t = nb::ndarray<double, nb::ndim<1>, nb::f_contig>;
using matrix_t = nb::ndarray<double, nb::ndim<2>, nb::f_contig>;

auto singular_values(matrix_t &matrix) {
    double *sv = new double[matrix.shape(1)];
    int info;

    if (matrix.shape(0) < matrix.shape(1)) {
        throw std::invalid_argument("Input matrix must not have more columns than rows");
    }

    compute_singular_values_python(static_cast<int>(matrix.shape(0)), static_cast<int>(matrix.shape(1)), static_cast<double *>(matrix.data()), 1, sv, &info);

    if (info == 1000) {
        throw std::runtime_error("MPI error");
    } else if (info > 0) {
        throw std::runtime_error("Zolo-SVD error");
    }

    nb::capsule owner(sv, [](void *p) noexcept {
        delete[] (double *) p;
    });

    return nb::ndarray<nb::numpy, double, nb::ndim<1>, nb::f_contig>(sv, {matrix.shape(1)}, owner);
}

auto singular_vectors(matrix_t &matrix) {
    double *sv = new double[matrix.shape(1)];
    double *left = new double[matrix.shape(0) * matrix.shape(1)];
    double *right = new double[matrix.shape(1) * matrix.shape(1)];
    int info;

    if (matrix.shape(0) < matrix.shape(1)) {
        throw std::invalid_argument("Input matrix must not have more columns than rows");
    }

    compute_singular_vectors_python(static_cast<int>(matrix.shape(0)), static_cast<int>(matrix.shape(1)), static_cast<double *>(matrix.data()), 1, sv, left, right, &info);

    if (info == 1000) {
        throw std::runtime_error("MPI error");
    } else if (info > 0) {
        throw std::runtime_error("Zolo-SVD error");
    }

    nb::capsule owner_sv(sv, [](void *p) noexcept {
        delete[] (double *) p;
    });
    nb::capsule owner_left(left, [](void *p) noexcept {
        delete[] (double *) p;
    });
    nb::capsule owner_right(right, [](void *p) noexcept {
        delete[] (double *) p;
    });

    return std::make_tuple(nb::ndarray<nb::numpy, double, nb::ndim<1>, nb::f_contig>(sv, {matrix.shape(1)}, owner_sv),
                           nb::ndarray<nb::numpy, double, nb::ndim<2>, nb::f_contig>(left, {matrix.shape(0), matrix.shape(1)}, owner_left),
                           nb::ndarray<nb::numpy, double, nb::ndim<2>, nb::f_contig>(right, {matrix.shape(1), matrix.shape(1)}, owner_right));
}

std::function<double(double)> invert_singular_values_handle;
double invert_singular_values_function(double sv) {
    return invert_singular_values_handle(sv);
}

auto solve_system(matrix_t &matrix, vector_t &rhs, std::function<double(double)> invert_singular_values) {
    double *solution = new double[matrix.shape(1)];
    int info;

    if (matrix.shape(0) < matrix.shape(1)) {
        throw std::invalid_argument("Input matrix must not have more columns than rows");
    }

    invert_singular_values_handle = invert_singular_values;
    solve_system_python(static_cast<int>(matrix.shape(0)), static_cast<int>(matrix.shape(1)), static_cast<double *>(matrix.data()), static_cast<double *>(rhs.data()), invert_singular_values_function, 1, solution, &info);
    invert_singular_values_handle = NULL;

    if (info == 1000) {
        throw std::runtime_error("MPI error");
    } else if (info > 0) {
        throw std::runtime_error("Zolo-SVD error");
    }

    nb::capsule owner(solution, [](void *p) noexcept {
        delete[] (double *) p;
    });

    return nb::ndarray<nb::numpy, double, nb::ndim<1>, nb::f_contig>(solution, {matrix.shape(1)}, owner);
}

NB_MODULE(libzolo_svd, m) {

    m.def("singular_values", &singular_values, R""""(
Computes the singular values of a given matrix

Parameters
----------
matrix : numpy.ndarray
    Input matrix with equal values on all MPI tasks

Returns
-------
sv : numpy.ndarray
    Array of singular values in increasing order

Raises
------
RuntimeError
    If MPI is not initialized or any other calculation error occurred)"""");

    m.def("singular_vectors", &singular_vectors, R""""(
Computes the singular values and singular vectors of a given matrix

Parameters
----------
matrix : numpy.ndarray
    Input matrix with equal values on all MPI tasks

Returns
-------
sv : numpy.ndarray
    Array of singular values in increasing order
u : numpy.ndarray
    Matrix of the left singular vectors
v : numpy.ndarray
    Matrix of the right singular vectors

Raises
------
RuntimeError
    If MPI is not initialized or any other calculation error occurred)"""");

    m.def("solve", &solve_system, R""""(
Solves the linear system 'matrix * x = b' using the singular value decomposition
to compute the pseudo-inverse matrix

Parameters
----------
matrix : numpy.ndarray
    Square or tall input matrix with equal values on all MPI tasks
b : numpy.ndarray
    Right-hand side vector
invert_singular_values : callback function sv -> f(sv)
    Callback function returning the inverse of a given singular value

Returns
-------
x : numpy.ndarray
    Solution vector

Raises
------
RuntimeError
    If MPI is not initialized or any other calculation error occurred)"""");

}
