module debug
  use, intrinsic :: iso_fortran_env, only: stderr => error_unit
  implicit none
  public

  type error_t
    private
    integer :: stack_pointer = 0
    integer :: info_stack(10)
    character(72) :: message_stack(10)

    contains
      procedure :: has_error => error_has_error
      procedure :: add_error => error_add_error
      procedure :: get_info => error_get_info
      procedure :: check_info => error_check_info
      procedure :: print => error_print
      procedure :: clear => error_clear
  end type

  integer, parameter :: ERROR_M = -3
  integer, parameter :: ERROR_N = -4
  integer, parameter :: ERROR_A = -5
  integer, parameter :: ERROR_IA = -6
  integer, parameter :: ERROR_JA = -7
  integer, parameter :: ERROR_DESCA = -8
  integer, parameter :: ERROR_U = -10
  integer, parameter :: ERROR_IU = -11
  integer, parameter :: ERROR_JU = -12
  integer, parameter :: ERROR_DESCU = -13
  integer, parameter :: ERROR_VT = -14
  integer, parameter :: ERROR_IVT = -15
  integer, parameter :: ERROR_JVT = -16
  integer, parameter :: ERROR_DESCVT = -17

  contains

    function error_has_error(this) result(has_error)
      class(error_t), intent(in) :: this
      logical :: has_error

      has_error = this%stack_pointer > 0
    end function

    subroutine error_add_error(this, message, info)
      class(error_t), intent(inout) :: this
      character(*), intent(in) :: message
      integer, intent(in), optional :: info

      this%stack_pointer = this%stack_pointer + 1
      this%message_stack(this%stack_pointer) = message
      if (present(info)) then
        this%info_stack(this%stack_pointer) = info
      else
        this%info_stack(this%stack_pointer) = 0
      end if
    end subroutine

    function error_get_info(this) result(info)
      class(error_t), intent(in) :: this
      integer :: info

      if (this%has_error()) then
        info = this%info_stack(this%stack_pointer)
        if (info == 0) then
          info = 1
        end if
      else
        info = 0
      end if
    end function

    function error_check_info(this, info, message) result(has_error)
      class(error_t), intent(inout) :: this
      integer, intent(in) :: info
      character(*), intent(in) :: message
      logical :: has_error

      if (info /= 0) then
        call this%add_error(message, info)
      end if
      has_error = this%has_error()
    end function

    subroutine error_print(this)
      class(error_t), intent(in) :: this
      integer :: i

      if (this%has_error()) then
        write(stderr,'(a)') "Error stack (most recent message last):"
        do i = this%stack_pointer, 1, -1
          if (this%info_stack(i) /= 0) then
            write(stderr,'(a,1x,a,a,1x,i3)') ">>", trim(this%message_stack(i)), ":", this%info_stack(i)
          else
            write(stderr,'(a,1x,a)') ">>", trim(this%message_stack(i))
          end if
        end do
      end if
    end subroutine

    subroutine error_clear(this)
      class(error_t), intent(inout) :: this

      this%stack_pointer = 0
    end subroutine

end module
