#:include 'common.fypp'
module zolo
  use auxil
  use constants
  use debug, only: error_t
  use ftimings_zsvd, only: timer_t
  use matrix
  implicit none

  private
#:for id in IDS
  public :: zolo_pd_${id}$
#:endfor

  contains

#:for id, dtype, prec in TYPES
#:set one = CONST(1.0, dtype, prec)
#:set zero = CONST(0.0, dtype, prec)
    !> This routine computes the polar decomposition of matrix A by using Zolo-PD.
    !> On exit, matrix Q stores the computed polar factors.
    subroutine zolo_pd_${id}$(matrix_a, matrix_q, rk, alpha, beta, err, timer)
      type(matrix_${id}$_t), intent(in) :: matrix_a
      type(matrix_${id}$_t), intent(inout) :: matrix_q
      integer, intent(in) :: rk
      real(${prec}$), intent(in) :: alpha, beta
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer

      call matrix_q%copy_from(matrix_a) ! Q := A

      call matrix_q%factor(${one}$, ${one}$ * beta, err) ! Q := Q / beta
      if (err%has_error()) then
        call err%add_error("Error during multiplication of matrix Q")
        return
      end if

      if (rk == 1) then
        call zolo_pd_1_${id}$(matrix_q, alpha, beta, err, timer)
      else
        call zolo_pd_r_${id}$(matrix_q, rk, alpha, beta, err, timer)
      end if
    end subroutine

    !> Polar decomposition (general case for any rk)
    !> This routine is the distributed parallel version with rk communicators.
    !> On entry, matrix Q is distributed over all processes.
    !>
    !> See Nakatsukasa & Freund, SIAM Review 58 (2016)
    !> and Li, Liu, & Du, Parallel Computing 86 (2019)
    subroutine zolo_pd_r_${id}$(matrix_q, rk, alpha, beta, err, timer)
      type(matrix_${id}$_t), intent(inout) :: matrix_q
      integer, intent(in) :: rk
      real(${prec}$), intent(in) :: alpha, beta
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer

      real(${prec}$) :: it_err, it_err_prev, con, tol, eps, fcon, Mhat
      real(${prec}$) :: cj(2*rk), aj(rk), max_coeff
      ${dtype}$(${prec}$), allocatable :: array_work(:)
      integer :: nblk, it, myid, nprocs, i, j, ip, npr_dummy, npc_dummy
      integer :: sep_nprows, sep_npcols, sep_nprocs, tp_nprows, tp_npcols, tp_myprow, tp_mypcol
      integer :: tp_context, context_r(rk), myprow, mypcol, sep_desc(desc_size)
      integer, allocatable :: umap(:,:)
      logical :: should_handle(rk)
      type(matrix_${id}$_t) :: matrix_x, matrix_tt, matrix_work_small, matrix_work_transpose, matrix_work_tall

      nblk = matrix_q%nblk

      call blacs_pinfo(myid, nprocs)

      ! **********************************************************
      !          Split the processes into rk groups              *
      ! **********************************************************
      sep_nprocs = nprocs / rk

      ! Create a new grid: (rk, sep_nprocs)
      call blacs_get(matrix_q%get_context(), 0, tp_context)
      tp_nprows = rk
      tp_npcols = sep_nprocs
      call blacs_gridinit(tp_context, 'R', tp_nprows, tp_npcols)
      call blacs_gridinfo(tp_context, tp_nprows, tp_npcols, tp_myprow, tp_mypcol)

      do sep_npcols = nint(sqrt(real(sep_nprocs))), 2, -1
        if (mod(sep_nprocs, sep_npcols) == 0) exit
      end do
      sep_nprows = sep_nprocs / sep_npcols

      ! Create new contexts for each rk
      allocate(umap(sep_nprows,sep_npcols))
      do ip = 1, rk
        do i = 1, sep_nprows
          do j = 1, sep_npcols
            umap(i,j) = (ip-1)*sep_nprocs + (i-1)*sep_npcols + (j-1)
          end do
        end do
        call blacs_get(tp_context, 0, context_r(ip))
        call blacs_gridmap(context_r(ip), umap(:,:), sep_nprows, sep_nprows, sep_npcols)
      end do
      deallocate(umap)

      !  sep_npcols ^
      !             |
      !             |
      !             |
      !             |--------->
      !            /      sep_nprows
      !           /
      !          /
      !         L  rk

      should_handle(1:rk) = .false.

      ! Allocate matrices in these sub-contexts
      do ip = 1, rk
        call blacs_gridinfo(context_r(ip), npr_dummy, npc_dummy, myprow, mypcol)
        if (myprow > -1 .and. mypcol > -1) then
          should_handle(ip) = .true.

          call matrix_x%alloc(context_r(ip), nblk, matrix_q%rows, matrix_q%cols, err)
          call matrix_x%set(${zero}$, ${zero}$)
          call matrix_tt%alloc_like(matrix_x, err)
          call matrix_work_small%alloc(context_r(ip), nblk, matrix_x%cols, matrix_x%cols, err)
          call matrix_work_transpose%alloc(context_r(ip), nblk, matrix_x%cols, matrix_x%rows, err)
          call matrix_work_tall%alloc(context_r(ip), nblk, matrix_x%rows+matrix_x%cols, matrix_x%cols, err)
        end if
      end do

      if (err%has_error()) then
        do ip = 1, rk
          if (should_handle(ip)) then
            call blacs_gridexit(context_r(ip))
          end if
        end do
        call blacs_gridexit(tp_context)
        return
      end if

      allocate(array_work(matrix_q%cols))

      ! Fetch the descriptor if on the rk=1 process
      if (should_handle(1)) then
        sep_desc(:) = matrix_x%desc(:)
      else
        sep_desc(:) = 0
        sep_desc(desc_context) = -1
      end if

      ! Copy the matrix Q to the matrix X on the rk=1 process
      call p${id}$gemr2d(matrix_q%rows, matrix_q%cols, matrix_q%values, 1, 1, matrix_q%desc, matrix_x%values, 1, 1, sep_desc, tp_context)

      ! Sum along the rk-axis of the tp_context == Distribute the matrix X from rk=1 process to all other rk processes
      do ip = 0, tp_npcols-1
        if (tp_mypcol == ip) then
          call ${id}$gsum2d(tp_context, 'COLUMN', ' ', matrix_x%na_rows, matrix_x%na_cols, matrix_x%values, matrix_x%na_rows, -1, ip)
        end if
      end do

      ! **********************************************************
      !                   Main iterative steps                   *
      ! **********************************************************

      eps = epsilon(0._${prec}$) * 0.5_${prec}$ * radix(0._${prec}$)
      tol = eps**(1._${prec}$ / (real(rk, kind=${prec}$) * 2._${prec}$ + 1._${prec}$))
      it = 0
      it_err = 1._${prec}$
      it_err_prev = 10._${prec}$
      con = alpha / beta

      call timer%start("Loop")

      main_loop: do while (it_err > tol .and. (it_err < it_err_prev .or. it < 2))

        ! Compute the coefficient cj and aj
        call zolo_coeff_${prec}$(con, eps, rk, cj, aj)
        max_coeff = maxval(cj(1:2*rk))

        if ((it == 0 .or. con > 1000._${prec}$ * alpha) .and. max_coeff > 100._${prec}$) then
          do ip = 1, rk
            if (should_handle(ip)) then
              call timer%start("Zolo-QR")
              call zolo_qr_${id}$(matrix_x, matrix_tt, matrix_work_tall, array_work, cj(2*ip-1), aj(ip), err, timer)
              call timer%stop("Zolo-QR")
              if (err%has_error()) exit main_loop
            end if
          end do

          do ip = 0, tp_npcols-1
            if (tp_mypcol == ip) then
              call timer%start("Sum Tj")
              call ${id}$gsum2d(tp_context, 'COLUMN', ' ', matrix_tt%na_rows, matrix_tt%na_cols, matrix_tt%values,&
                  & matrix_tt%na_rows, -1, ip) ! sum of Tj
              call timer%stop("Sum Tj")
            end if
          end do
        else
          do ip = 1, rk
            if (should_handle(ip)) then
              call timer%start("Zolo-Chol")
              call zolo_chol_${id}$(matrix_x, matrix_tt, matrix_work_small, matrix_work_transpose, cj(2*ip-1), aj(ip), err, timer)
              call timer%stop("Zolo-Chol")
              if (err%has_error()) exit main_loop
            end if
          end do

          do ip = 0, tp_npcols-1
            if (tp_mypcol == ip) then
              call timer%start("Sum Tj")
              call ${id}$gsum2d(tp_context, 'COLUMN', ' ', matrix_tt%na_rows, matrix_tt%na_cols, matrix_tt%values,&
                  & matrix_tt%na_rows, -1, ip) ! sum of Tj
              call timer%stop("Sum Tj")
            end if
          end do
        end if

        call timer%start("Form X2")
        Mhat = 1._${prec}$
        do i = 1, rk
          Mhat = Mhat * (1._${prec}$ + cj(2*i-1)) / (1._${prec}$ + cj(2*i))
        end do

        fcon = 1._${prec}$
        do i = 1, rk
          fcon = fcon * (con**2 + cj(2*i)) / (con**2 + cj(2*i-1))
        end do
        con = max(fcon * con * Mhat, 1._${prec}$)
        it_err_prev = it_err

        do ip = 1, rk
          if (should_handle(ip)) then
            call form_x2_${id}$(matrix_x, matrix_tt, Mhat, it_err, err)
            if (err%has_error()) exit main_loop
          end if
        end do
        call timer%stop("Form X2")

        it = it + 1
      end do main_loop

      call timer%stop("Loop")

      ! Copy the matrix X from rk=1 process back to Q
      call p${id}$gemr2d(matrix_x%rows, matrix_x%cols, matrix_x%values, 1, 1, sep_desc, matrix_q%values, 1, 1, matrix_q%desc, tp_context)

      do ip = 1, rk
        if (should_handle(ip)) then
          call blacs_gridexit(context_r(ip))
        end if
      end do
      call blacs_gridexit(tp_context)

      if (err%has_error()) then
        call err%add_error("Error in Zolo-PD loop")
        return
      end if
    end subroutine

    !> Polar decomposition (rk=1 case)
    !> This routine is the version without further sub-parallelization.
    subroutine zolo_pd_1_${id}$(matrix_q, alpha, beta, err, timer)
      type(matrix_${id}$_t), intent(inout) :: matrix_q
      real(${prec}$), intent(in) :: alpha, beta
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer

      real(${prec}$) :: it_err, it_err_prev, con, tol, eps, fcon, Mhat
      real(${prec}$) :: cj(2), aj, max_coeff
      ${dtype}$(${prec}$), allocatable :: array_work(:)
      integer :: it
      type(matrix_${id}$_t) :: matrix_tt, matrix_work_small, matrix_work_transpose, matrix_work_tall

      ! Allocate matrices
      call matrix_tt%alloc_like(matrix_q, err)
      call matrix_work_small%alloc_like(matrix_q, err, matrix_q%cols, matrix_q%cols)
      call matrix_work_transpose%alloc_like(matrix_q, err, matrix_q%cols, matrix_q%rows)
      call matrix_work_tall%alloc_like(matrix_q, err, matrix_q%rows+matrix_q%cols, matrix_q%cols)

      if (err%has_error()) then
        return
      end if

      call matrix_tt%copy_from(matrix_q)

      allocate(array_work(matrix_q%cols))

      ! **********************************************************
      !                   Main iterative steps                   *
      ! **********************************************************

      eps = epsilon(0._${prec}$) * 0.5_${prec}$ * radix(0._${prec}$)
      tol = eps**(1._${prec}$ / 3._${prec}$)
      it = 0
      it_err = 1._${prec}$
      it_err_prev = 10._${prec}$
      con = alpha / beta

      call timer%start("Loop")

      main_loop: do while (it_err > tol .and. (it_err < it_err_prev .or. it < 2))

        ! Compute the coefficient cj and aj
        call zolo_coeff_${prec}$(con, eps, cj, aj)
        max_coeff = maxval(cj(:))

        Mhat = (1._${prec}$ + cj(1)) / (1._${prec}$ + cj(2))

        fcon = (con**2 + cj(2)) / (con**2 + cj(1))
        con = max(fcon * con * Mhat, 1._${prec}$)
        it_err_prev = it_err

        if ((it == 0 .or. con > 1000._${prec}$ * alpha) .and. max_coeff > 100._${prec}$) then
          call timer%start("Zolo-QR")
          call zolo_qr_${id}$(matrix_q, matrix_tt, matrix_work_tall, array_work, cj(1), aj, err, timer, Mhat, it_err)
          call timer%stop("Zolo-QR")
          if (err%has_error()) exit main_loop
        else
          call timer%start("Zolo-Chol")
          call zolo_chol_${id}$(matrix_q, matrix_tt, matrix_work_small, matrix_work_transpose, cj(1), aj, err, timer, Mhat, it_err)
          call timer%stop("Zolo-Chol")
          if (err%has_error()) exit main_loop
        end if
        it = it + 1
      end do main_loop

      call timer%stop("Loop")

      if (err%has_error()) then
        call err%add_error("Error in Zolo-PD loop")
        return
      end if
    end subroutine

    !> QR factorization
    !>
    !> See Li et al. (2019), Algorithm 3, Step 4.(c), upper part
    subroutine zolo_qr_${id}$(matrix_x1, matrix_tt, matrix_work, array_work, cj, aj, err, timer, Mhat, it_err)
      type(matrix_${id}$_t), intent(inout) :: matrix_x1, matrix_tt, matrix_work
      ${dtype}$(${prec}$), intent(inout) :: array_work(:)
      real(${prec}$), intent(in) :: cj, aj
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer
      real(${prec}$), intent(in), optional :: Mhat
      real(${prec}$), intent(out), optional :: it_err

      integer :: info, lwork
      ${dtype}$(${prec}$) :: sqrt_cj, aj_sqrt_cj, worksize(1)
      ${dtype}$(${prec}$), allocatable :: work(:)

      call matrix_work%copy_from(matrix_x1) ! work(upper part) := X1

#:if dtype == 'real'
      sqrt_cj = sqrt(cj)
#:else
      sqrt_cj = cmplx(sqrt(cj), kind=${prec}$)
#:endif
      call p${id}$laset('A', matrix_x1%cols, matrix_x1%cols, ${zero}$, sqrt_cj,&
          & matrix_work%values, matrix_x1%rows+1, 1, matrix_work%desc) ! work(lower part) := sqrt(c)*I

      call timer%start("mp${id}$geqrf")
      call mp${id}$geqrf(matrix_work%rows, matrix_work%cols, matrix_work%values, 1, 1, matrix_work%desc, array_work, worksize, -1, info)
      if (err%check_info(info, "Error in mp${id}$geqrf")) return
      lwork = int(worksize(1)) + 1

      allocate(work(lwork), source=${zero}$)

      call mp${id}$geqrf(matrix_work%rows, matrix_work%cols, matrix_work%values, 1, 1, matrix_work%desc, array_work, work, lwork, info)
      if (err%check_info(info, "Error in mp${id}$geqrf")) return
      call timer%stop("mp${id}$geqrf")

      call timer%start("mp${id}$orgqr")
      call mp${id}$orgqr(matrix_work%rows, matrix_work%cols, matrix_work%cols, matrix_work%values, 1, 1, matrix_work%desc,&
          & array_work, work, lwork, info)
      if (err%check_info(info, "Error in mp${id}$orgqr")) return
      call timer%stop("mp${id}$orgqr")

#:if dtype == 'real'
      aj_sqrt_cj = aj / sqrt_cj
#:else
      aj_sqrt_cj = cmplx(aj, kind=${prec}$) / sqrt_cj
#:endif

      if (present(Mhat)) then ! rk=1 case
        call timer%start("p${id}$gemm")
        call p${id}$gemm('N', 'C', matrix_tt%rows, matrix_tt%cols, matrix_work%cols, aj_sqrt_cj * Mhat, matrix_work%values, 1, 1,&
            & matrix_work%desc, matrix_work%values, matrix_x1%rows+1, 1, matrix_work%desc, ${one}$ * Mhat,&
            & matrix_tt%values, 1, 1, matrix_tt%desc) ! T := Mhat * (aj_sqrt_cj * work(upper part) * work(lower part)^T + X1)
        call timer%stop("p${id}$gemm")

        call matrix_x1%sub(matrix_tt) ! X1 := X1 - T = X1 - X2

        it_err = matrix_x1%norm() / matrix_tt%norm() ! Frobenius norm of X1 - X2 / Frobenius norm of X2

        call matrix_x1%copy_from(matrix_tt) ! X1 := X2
      else
        call timer%start("p${id}$gemm")
        call p${id}$gemm('N', 'C', matrix_tt%rows, matrix_tt%cols, matrix_work%cols, aj_sqrt_cj, matrix_work%values, 1, 1,&
            & matrix_work%desc, matrix_work%values, matrix_x1%rows+1, 1, matrix_work%desc, ${zero}$,&
            & matrix_tt%values, 1, 1, matrix_tt%desc) ! T := aj_sqrt_cj * work(upper part) * work(lower part)^T
        call timer%stop("p${id}$gemm")
      end if
    end subroutine

    !> Cholesky factorization
    !>
    !> See Li et al. (2019), Algorithm 3, Step 4.(c), lower part
    subroutine zolo_chol_${id}$(matrix_x1, matrix_tt, matrix_work_x, matrix_work_t, cj, aj, err, timer, Mhat, it_err)
      type(matrix_${id}$_t), intent(inout) :: matrix_x1, matrix_tt, matrix_work_x, matrix_work_t
      real(${prec}$), intent(in) :: cj, aj
      type(error_t), intent(inout) :: err
      type(timer_t), intent(inout) :: timer
      real(${prec}$), intent(in), optional :: Mhat
      real(${prec}$), intent(out), optional :: it_err

      integer :: info
      ${dtype}$(${prec}$) :: cj_local, aj_local

#:if dtype == 'real'
      cj_local = cj
      aj_local = aj
#:else
      cj_local = cmplx(cj, kind=${prec}$)
      aj_local = cmplx(aj, kind=${prec}$)
#:endif

      call matrix_work_x%set(cj_local, ${zero}$) ! cj*I
      call timer%start("transp_mm")
      call matrix_work_x%transp_mm(matrix_x1, add=.true.) ! work_x := X1^T * X1 + cj*I
      call timer%stop("transp_mm")

      ! work_x is now Z_{2j-1}

      call matrix_work_t%transp(matrix_x1) ! work_t := X1^T

      call timer%start("p${id}$posv")
      call p${id}$posv('U', matrix_work_x%rows, matrix_work_t%cols, matrix_work_x%values, 1, 1, matrix_work_x%desc,&
          & matrix_work_t%values, 1, 1, matrix_work_t%desc, info) ! solve: work_x * Y = work_t for Y and set work_t := Y
      call timer%stop("p${id}$posv")
      if (err%check_info(info, "Error in p${id}$posv")) return

      if (present(Mhat)) then ! rk=1 case
        call matrix_tt%transp(matrix_work_t, aj_local * Mhat, ${one}$ * Mhat) ! T := Mhat * (aj * work_t^T + T)

        ! T is now X2
        call matrix_x1%sub(matrix_tt) ! X1 := X1 - T = X1 - X2

        it_err = matrix_x1%norm() / matrix_tt%norm() ! Frobenius norm of X1 - X2 / Frobenius norm of X2

        call matrix_x1%copy_from(matrix_tt) ! X1 := X2
      else
        call matrix_tt%transp(matrix_work_t, aj_local) ! T := aj * work_t^T
      end if
    end subroutine

    !> Computes X2 for the next iteration and stores it in matrix_x1 on exit.
    !> The variable err contains the Frobenius norm of X2-X1.
    !>
    !> See Li et al. (2019), Algorithm 3, Step 4.(d)
    subroutine form_x2_${id}$(matrix_x1, matrix_t_sum, Mhat, it_err, err)
      type(matrix_${id}$_t), intent(inout) :: matrix_x1, matrix_t_sum
      real(${prec}$), intent(in) :: Mhat
      real(${prec}$), intent(out) :: it_err
      type(error_t), intent(inout) :: err

      call matrix_t_sum%add(matrix_x1) ! T := X1 + sum(T)
      call matrix_t_sum%factor(${one}$ * Mhat, ${one}$, err) ! T := T * Mhat
      if (err%has_error()) then
        call err%add_error("Error in Form X2")
        return
      end if

      ! T is now X2

      call matrix_x1%sub(matrix_t_sum) ! X1 := X1 - T = X1 - X2

      it_err = matrix_x1%norm() / matrix_t_sum%norm() ! Frobenius norm of X1 - X2 / Frobenius norm of X2

      call matrix_x1%copy_from(matrix_t_sum) ! X1 := X2
    end subroutine
#:endfor

end module
