#:include 'common.fypp'
module zolo_svd_scalapack_api
  use constants, only: sp, dp, desc_size
  use debug
  use matrix, only: matrix_size
  use zolo_svd
  implicit none

  private
#:for id in IDS
  public p${id}$gesvd_zolo
#:endfor

  contains

#:for id, dtype, prec in TYPES
#:set one = CONST(1.0, dtype, prec)
#:set zero = CONST(0.0, dtype, prec)
    !> Compute the singular value decomposition (SVD) using the Zolo-SVD algorithm.
    !> This interface is similar to ScaLAPACK's p?gesvd (see, e.g., https://manpages.org/pdgesvd) and can be used
    !> as a drop-in replacement of the latter.
    !>
    !> Difference to ScaLAPACK's p?gesvd: If you provide lwork == 1, then work(1) will contain the accuracy of the SVD on exit.
    subroutine p${id}$gesvd_zolo(jobu, jobvt, m, n, a, ia, ja, desca, s, u, iu, ju, descu, vt, ivt, jvt, descvt, work, lwork,&
#:if dtype == 'complex'
        & rwork,&
#:endif
        & info)
      character, intent(in) :: jobu, jobvt
      integer, intent(in) :: m, n
      ${dtype}$(${prec}$), target, intent(in) :: a(*)
      integer, intent(in) :: ia, ja
      integer, intent(in) :: desca(*)
      real(${prec}$), intent(inout) :: s(*)
      ${dtype}$(${prec}$), target, intent(inout) :: u(*)
      integer, intent(in) :: iu, ju
      integer, intent(in) :: descu(*)
      ${dtype}$(${prec}$), target, intent(inout) :: vt(*)
      integer, intent(in) :: ivt, jvt
      integer, intent(in) :: descvt(*)
      ${dtype}$(${prec}$), intent(inout) :: work(*)
      integer, intent(in) :: lwork
#:if dtype == 'complex'
      real(${prec}$), intent(inout) :: rwork(*)
#:endif
      integer, intent(out) :: info

      ${dtype}$(${prec}$), pointer :: a_p(:,:), u_p(:,:), vt_p(:,:)
      ${dtype}$(${prec}$) :: fail_fast
      integer :: ma, na, mu, nu, mvt, nvt
      logical :: compute_left, compute_right

      type(zolo_svd_t) :: this

      if (lwork == -1) then
        work(1) = ${zero}$
#:if dtype == 'complex'
        rwork(1) = 0._${prec}$
#:endif
        info = 0
        return
      end if

      if (m /= n) then
        info = ERROR_M
        return
      end if

      if (ia /= 1) then
        info = ERROR_IA
        return
      end if
      if (ja /= 1) then
        info = ERROR_JA
        return
      end if
      if (iu /= 1) then
        info = ERROR_IU
        return
      end if
      if (ju /= 1) then
        info = ERROR_JU
        return
      end if
      if (ivt /= 1) then
        info = ERROR_IVT
        return
      end if
      if (jvt /= 1) then
        info = ERROR_JVT
        return
      end if

      if (any(descu(1:desc_size) /= desca(1:desc_size))) then
        info = ERROR_DESCU
        return
      end if
      if (any(descvt(1:desc_size) /= desca(1:desc_size))) then
        info = ERROR_DESCVT
        return
      end if

      compute_left = jobu == 'v' .or. jobu == 'V'
      compute_right = jobvt == 'v' .or. jobvt == 'V'

      call this%set_option(ZSVD_COMPUTE_LEFT_VECTORS, convert_bool(compute_left))
      call this%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, convert_bool(compute_right))
      call this%set_option(ZSVD_DECREASING_ORDER, 1)
      call this%set_option(ZSVD_COMPUTE_ACCURACY, convert_bool(lwork == 1 .and. compute_left .and. compute_right))

      if (compute_left .and. compute_right) then
        call this%set_descriptors(desca, descu, descvt)
      else if (compute_left) then
        call this%set_descriptors(desca, descu, descvt)
      else if (compute_right) then
        call this%set_descriptors(desca, desc_right_vectors=descvt)
      else
        call this%set_descriptors(desca)
      end if

      call matrix_size(desca, ma, na)
      if (ma == 0 .or. na == 0) then
        info = ERROR_DESCA
        return
      end if
      a_p(1:ma,1:na) => a(1:ma*na)
      ! Segfaults here if the user did not provide correct arrays
      fail_fast = a_p(ma,na)

      if (compute_left) then
        call matrix_size(descu, mu, nu)
        if (mu == 0 .or. nu == 0) then
          info = ERROR_DESCU
          return
        end if
        u_p(1:mu,1:nu) => u(1:mu*nu)
        fail_fast = u_p(mu,nu)
      end if

      if (compute_left .or. compute_right) then
        call matrix_size(descvt, mvt, nvt)
        if (mvt == 0 .or. nvt == 0) then
          info = ERROR_DESCVT
          return
        end if
        vt_p(1:mvt,1:nvt) => vt(1:mvt*nvt)
        fail_fast = vt_p(mvt,nvt)
      end if

      if (compute_left) then
        info = this%compute_svd(a_p, s(1:n), u_p, vt_p)
      else if (compute_right) then
        info = this%compute_svd(a_p, s(1:n), right_singular_vectors=vt_p)
      else
        info = this%compute_svd(a_p, s(1:n))
      end if

      nullify(a_p)
      nullify(u_p)
      nullify(vt_p)

      if (info /= 0) then
        return
      end if

      if (compute_right) then
        ! Transpose right matrix to be consistent with ScaLAPACK API
        call p${id}$lacpy('F', n, n, vt, 1, 1, descvt, a, 1, 1, desca)
#:if dtype == 'complex'
        call p${id}$tranu(n, n, ${one}$, a, 1, 1, desca, ${zero}$, vt, 1, 1, descvt)
#:else
        call p${id}$tran(n, n, ${one}$, a, 1, 1, desca, ${zero}$, vt, 1, 1, descvt)
#:endif
      end if

      if (lwork == 1) then
        work(1) = ${one}$ * real(this%get_accuracy(), kind=${prec}$)
      end if
    end subroutine
#:endfor

    function convert_bool(val)
      logical, intent(in) :: val
      integer :: convert_bool

      if (val) then
        convert_bool = 1
      else
        convert_bool = 0
      end if
    end function

end module
