#:def ASSERT(condition, message)
if (.not. (${condition}$)) then
  call err%add_error(${message}$)
  call err%print()
  call blacs_gridexit(context)
  call mpi_abort(MPI_COMM_WORLD, 1, ierr)
end if
#:enddef

#:def ASSERT_OK()
if (err%has_error()) then
  call err%print()
  call blacs_gridexit(context)
  call mpi_abort(MPI_COMM_WORLD, 1, ierr)
end if
#:enddef

program unittest_matrix
  use constants, only: dp
  use debug, only: error_t
  use matrix, only: matrix_d_t, matrix_size
#:if defined('WITH_MPI_MOD')
  use mpi
#:endif
  implicit none
#:if not defined('WITH_MPI_MOD')
  include 'mpif.h'
#:endif

  integer :: nblk = 4, rows = 7, cols = 9
  integer :: ierr, myid, nprocs, context, np_cols, np_rows, i, j
  real(dp) :: val, norm1
  type(error_t) :: err
  type(matrix_d_t) :: matrix_a, matrix_b, matrix_c
  real(dp), allocatable :: array(:,:), diag(:)
  integer, allocatable :: local_rows(:), local_cols(:)
  integer, allocatable :: global_rows(:), global_cols(:)
  integer, allocatable :: idrows(:), idcols(:)

  real(dp), external :: pdlange

  call mpi_init(ierr)
  call mpi_comm_rank(MPI_COMM_WORLD, myid, ierr)
  call mpi_comm_size(MPI_COMM_WORLD, nprocs, ierr)

  call blacs_pinfo(myid, nprocs)

  do np_cols = nint(sqrt(real(nprocs))), 2, -1
    if (mod(nprocs, np_cols) == 0) exit
  end do
  np_rows = nprocs / np_cols

  if (myid == 0) then
    write(*,'(2(a,i0))') "Matrix size=", rows, "x", cols
    write(*,'(3(a,i0))') "Number of processors: rows=", np_rows, ", cols=", np_cols, ", total=", nprocs
  end if

  call blacs_get(-1, 0, context)
  call blacs_gridinit(context, 'R', np_rows, np_cols)

  call matrix_a%alloc(context, nblk, rows, cols, err)
  @:ASSERT_OK()
  @:ASSERT(matrix_a%is_allocated, "Matrix A not allocated")

  call matrix_b%alloc_like(matrix_a, err)
  @:ASSERT_OK()

  call matrix_c%alloc_like(matrix_a, err, rows, rows)
  @:ASSERT_OK()

  call matrix_size(matrix_a%desc, i, j)
  @:ASSERT(i == matrix_a%na_rows .and. j == matrix_a%na_cols, "Inconsistent descriptor of matrix A")

  do i = 1, rows
    do j = 1, cols
      val = real((i-1) * cols + j, kind=dp)
      call matrix_a%set_at(i, j, val)
      call matrix_b%set_at(i, cols-j+1, val)
    end do
  end do

  allocate(local_rows(matrix_a%rows), idrows(matrix_a%rows))
  allocate(local_cols(matrix_a%cols), idcols(matrix_a%cols))
  allocate(global_rows(matrix_a%na_rows), global_cols(matrix_a%na_cols))
  call matrix_a%infog2l(local_rows, local_cols, idrows, idcols)
  call matrix_a%indxl2g(global_rows, global_cols)

  do i = 1, matrix_a%na_rows
    @:ASSERT(local_rows(global_rows(i)) == i, "Incorrect infog2l vs. indxl2g")
  end do
  do j = 1, matrix_a%na_cols
    @:ASSERT(local_cols(global_cols(j)) == j, "Incorrect infog2l vs. indxl2g")
  end do

  deallocate(local_rows, idrows)
  deallocate(local_cols, idcols)
  deallocate(global_rows, global_cols)

  call matrix_b%flip_cols(err)
  @:ASSERT_OK()
  call matrix_b%sub(matrix_a)
  @:ASSERT(matrix_b%norm() < 1.e-13_dp, "Incorrect norm of B after column flip")

  allocate(array(rows,cols))
  call matrix_a%copy_to_array(array)

  do i = 1, rows
    do j = 1, cols
      call matrix_b%set_at(i, j, array(i,j))
    end do
  end do

  call matrix_b%sub(matrix_a)
  @:ASSERT(matrix_b%norm() < 1.e-13_dp, "Incorrect norm of B after copy to array")

  call matrix_b%copy_from(matrix_a)
  call matrix_a%copy_from_array(array)

  call matrix_b%sub(matrix_a)
  @:ASSERT(matrix_b%norm() < 1.e-13_dp, "Incorrect norm of B after copy from array")

  call matrix_a%factor(4.0_dp, 2.0_dp, err)
  @:ASSERT_OK()

  call matrix_b%set(0.0_dp, 0.0_dp)
  call matrix_b%copy_from(matrix_a)
  @:ASSERT(abs(matrix_a%norm() - matrix_b%norm()) < 1.e-13_dp, "Different norms of A and B after copy_from")

  call matrix_c%set(1.0_dp, 0.0_dp)
  @:ASSERT(abs(matrix_c%norm()**2 - rows) < 1.e-13_dp, "Incorrect norm of C after set")

  call matrix_b%add(matrix_b)
  call matrix_b%sub(matrix_b)
  call matrix_b%sub(matrix_b)
  @:ASSERT(matrix_b%norm() < 1.e-13_dp, "Incorrect norm of B after add and sub")

  call matrix_b%set(0.0_dp, 0.0_dp)
  allocate(diag(min(rows,cols)))
  do i = 1, rows
    do j = 1, cols
      if (i == j) then
        array(i,j) = real(i, kind=dp)
        diag(i) = array(i,j)
      else
        array(i,j) = 0._dp
      end if
    end do
  end do
  call matrix_b%set_diag(diag)
  call matrix_a%copy_from_array(array)
  call matrix_b%sub(matrix_a)
  @:ASSERT(matrix_b%norm() < 1.e-13_dp, "Incorrect norm of B after set_diag")

  call matrix_b%dealloc()
  call matrix_b%alloc_like(matrix_a, err, cols, rows)
  call matrix_b%transp(matrix_a)
  call matrix_c%mm(matrix_a, matrix_b)
  norm1 = matrix_c%norm()
  call matrix_c%mm_transp(matrix_a, matrix_a)
  @:ASSERT(abs(norm1 - matrix_c%norm()) < 1.e-13_dp, "Different norms after transp/mm/mm_transp")

  call matrix_c%mm_transp(matrix_a)
  @:ASSERT(abs(norm1 - matrix_c%norm()) < 1.e-13_dp, "Different norms after mm_transp")

  call matrix_c%dealloc()
  call matrix_c%alloc_like(matrix_a, err, cols, cols)
  call matrix_c%mm(matrix_b, matrix_a)
  norm1 = matrix_c%norm()
  call matrix_c%transp_mm(matrix_a, matrix_a)
  @:ASSERT(abs(norm1 - matrix_c%norm()) < 1.e-13_dp, "Different norms after transp/mm/transp_mm")

  call matrix_c%transp_mm(matrix_a)
  @:ASSERT(abs(norm1 - matrix_c%norm()) < 1.e-13_dp, "Different norms after transp_mm")

  call matrix_a%dealloc()
  call matrix_b%dealloc()
  call matrix_c%dealloc()
  deallocate(array)
  deallocate(diag)

  if (myid == 0) then
    write(*,'(a)') "SUCCESS"
  end if

  call blacs_gridexit(context)
  call mpi_finalize(ierr)

end program
