program test_zolo_svd_performance
  use constants, only: dp
  use debug, only: error_t
  use matrix, only: matrix_d_t
#:if defined('WITH_MPI_MOD')
  use mpi
#:endif
  use zolo_svd
  implicit none
#:if not defined('WITH_MPI_MOD')
  include 'mpif.h'
#:endif

  integer :: ierr, myid, nprocs, num_args, rows, cols
  character(len=16) :: rows_str, cols_str

  call mpi_init(ierr)
  call mpi_comm_rank(MPI_COMM_WORLD, myid, ierr)
  call mpi_comm_size(MPI_COMM_WORLD, nprocs, ierr)

  num_args = command_argument_count()
  if (num_args == 2) then
    call get_command_argument(1, rows_str)
    call get_command_argument(2, cols_str)
    read(rows_str, '(I16)') rows
    read(cols_str, '(I16)') cols
  else
    rows = 10000
    cols = 8000
  end if

  call run_tests(myid, nprocs, rows, cols, 32)

  call mpi_finalize(ierr)

  contains

    subroutine run_tests(myid, nprocs, rows, cols, nblk)
      integer, intent(in) :: myid, nprocs, rows, cols, nblk

      integer :: context, np_rows, np_cols
      type(matrix_d_t) :: matrix_a

      do np_cols = nint(sqrt(real(nprocs))), 2, -1
        if (mod(nprocs, np_cols) == 0) exit
      end do
      np_rows = nprocs / np_cols

      if (myid == 0) then
        write(*,'(2(a,i0))') "Matrix size=", rows, "x", cols
        write(*,'(3(a,i0))') "Number of processors: rows=", np_rows, ", cols=", np_cols, ", total=", nprocs
      end if

      call blacs_get(-1, 0, context)
      call blacs_gridinit(context, 'R', np_rows, np_cols)

      call init_matrix(context, nblk, rows, cols, matrix_a)

      call test_with_vectors(matrix_a, 1, 2)

      call blacs_gridexit(context)
    end subroutine

    subroutine init_matrix(context, nblk, rows, cols, matrix_a)
      integer, intent(in) :: context, nblk, rows, cols
      type(matrix_d_t), intent(out) :: matrix_a

      integer :: i, j, info
      real(dp) :: rand
      type(error_t) :: err

      call matrix_a%alloc(context, nblk, rows, cols, err)
      if (err%has_error()) then
        call err%print()
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call random_seed()

      do i = 1, rows
        do j = 1, cols
          call random_number(rand)
          call matrix_a%set_at(i, j, rand)
        end do
     end do
    end subroutine

    subroutine test_with_vectors(matrix_a, rk, elpa, lower, upper)
      type(matrix_d_t), intent(in) :: matrix_a
      integer, intent(in) :: rk, elpa
      real(dp), intent(in), optional :: lower, upper

      integer :: info
      real(dp), allocatable :: singular_values(:)
      type(zolo_svd_t) :: svd
      type(matrix_d_t) :: matrix_left, matrix_right
      type(error_t) :: err

      allocate(singular_values(matrix_a%cols))

      call matrix_left%alloc_like(matrix_a, err)
      call matrix_right%alloc_like(matrix_a, err, matrix_a%cols, matrix_a%cols)
      if (err%has_error()) then
        call err%print()
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (present(lower) .and. present(upper)) then
        call svd%set_bounds(lower, upper)
      end if
      call svd%set_descriptors(matrix_a%desc, matrix_left%desc, matrix_right%desc)
      call svd%set_option(ZSVD_R, rk)
      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 1)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 1)
      call svd%set_option(ZSVD_COMPUTE_ACCURACY, 0)
      call svd%set_option(ZSVD_DECREASING_ORDER, 0)
      call svd%set_option(ZSVD_PRINT_TIMINGS, 1)
      call svd%set_option(ZSVD_ELPA_SOLVER, elpa)

      info = svd%compute_svd(matrix_a%values, singular_values, matrix_left%values, matrix_right%values)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if
    end subroutine

end program
