#:include '../src/common.fypp'
program test_zolo_svd_modern_api
  use constants, only: sp, dp
  use debug, only: error_t
  use matrix
#:if defined('WITH_MPI_MOD')
  use mpi
#:endif
  use zolo_svd
  implicit none
#:if not defined('WITH_MPI_MOD')
  include 'mpif.h'
#:endif

  integer :: ierr, myid, nprocs

  call mpi_init(ierr)
  call mpi_comm_rank(MPI_COMM_WORLD, myid, ierr)
  call mpi_comm_size(MPI_COMM_WORLD, nprocs, ierr)

#:for id in IDS
  call run_tests_${id}$(myid, nprocs, 17, 7, 4)
  call run_tests_${id}$(myid, nprocs, 200, 200, 16)
#:endfor

  if (myid == 0) then
    write(*,'(a)') "SUCCESS"
  end if

  call mpi_finalize(ierr)

  contains

#:for id, dtype, prec in TYPES
#:set one = CONST(1.0, dtype, prec)
#:set zero = CONST(0.0, dtype, prec)
    subroutine run_tests_${id}$(myid, nprocs, rows, cols, nblk)
      integer, intent(in) :: myid, nprocs, rows, cols, nblk

      integer :: context, np_rows, np_cols
      real(${prec}$), parameter :: sv_min = 17._${prec}$, sv_max = 27._${prec}$
      real(${prec}$) :: sv_min_bound, sv_max_bound
      type(matrix_${id}$_t) :: matrix_a

      do np_cols = nint(sqrt(real(nprocs))), 2, -1
        if (mod(nprocs, np_cols) == 0) exit
      end do
      np_rows = nprocs / np_cols

      if (myid == 0) then
        write(*,'(a)') "Type=${dtype}$, Precision=${prec}$"
        write(*,'(2(a,i0))') "Matrix size=", rows, "x", cols
        write(*,'(3(a,i0))') "Number of processors: rows=", np_rows, ", cols=", np_cols, ", total=", nprocs
      end if

      call blacs_get(-1, 0, context)
      call blacs_gridinit(context, 'R', np_rows, np_cols)

      call prepare_svd_${id}$(context, nblk, rows, cols, sv_min, sv_max, matrix_a)

      sv_min_bound = sv_min - 1._${prec}$
      sv_max_bound = sv_max + 1._${prec}$

      call test_without_vectors_${id}$(myid, matrix_a, 1, 1, sv_min, sv_max, sv_min_bound, sv_max_bound)
      call test_only_left_vectors_${id}$(matrix_a, 1, 1, sv_min_bound, sv_max_bound)
      call test_only_right_vectors_${id}$(matrix_a, 1, 1, sv_min_bound, sv_max_bound)
      call test_with_vectors_${id}$(myid, matrix_a, 1, 1, sv_min_bound, sv_max_bound)

      call test_without_vectors_${id}$(myid, matrix_a, 1, 2, sv_min, sv_max, sv_min_bound, sv_max_bound)
      call test_with_vectors_${id}$(myid, matrix_a, 1, 2, sv_min_bound, sv_max_bound)

      if (nprocs > 1) then
        call test_without_vectors_${id}$(myid, matrix_a, nprocs, 1, sv_min, sv_max, sv_min_bound, sv_max_bound)
        call test_with_vectors_${id}$(myid, matrix_a, nprocs, 1, sv_min_bound, sv_max_bound)
      end if

      ! Run some tests for really bad estimates of the bounds
      call test_without_vectors_${id}$(myid, matrix_a, 1, 1, sv_min, sv_max, 100._${prec}$, 100._${prec}$)
      call test_without_vectors_${id}$(myid, matrix_a, 1, 1, sv_min, sv_max, 1._${prec}$, 1._${prec}$)
      call test_without_vectors_${id}$(myid, matrix_a, 1, 1, sv_min, sv_max, 19._${prec}$, 23._${prec}$)

      ! Test without bounds estimates
      call test_without_vectors_${id}$(myid, matrix_a, 1, 1, sv_min, sv_max)
      call test_with_vectors_${id}$(myid, matrix_a, 1, 1)

      call test_invalid_${id}$(matrix_a)

      call test_solve_system_${id}$(matrix_a)

      call blacs_gridexit(context)
    end subroutine

    !> Define singular values and construct the matrix using LAPACK's ?latms
    subroutine prepare_svd_${id}$(context, nblk, rows, cols, sv_min, sv_max, matrix_a)
      integer, intent(in) :: context, nblk, rows, cols
      real(${prec}$), intent(in) :: sv_min, sv_max
      type(matrix_${id}$_t), intent(out) :: matrix_a

      integer :: i, j, iseed(4), info
      real(${prec}$), allocatable :: singular_values(:)
      ${dtype}$(${prec}$), allocatable :: work(:), local_matrix(:,:)
      type(error_t) :: err

      call matrix_a%alloc(context, nblk, rows, cols, err)
      if (err%has_error()) then
        call err%print()
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      allocate(singular_values(cols), work(4*cols), local_matrix(rows, cols))

      do i = 1, cols
        singular_values(i) = sv_min + ((sv_max - sv_min) / real(cols-1, kind=${prec}$)) * real(cols-i, kind=${prec}$)
      end do

      iseed(:) = 0
      call ${id}$latms(rows, cols, 'N', iseed, 'N', singular_values, 0, 0._${prec}$, 0._${prec}$, rows, cols, 'N', &
          local_matrix, rows, work, info)
      if (info /= 0) then
        write(*,'(a)') "${id}$latms error"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      do i = 1, rows
        do j = 1, cols
          call matrix_a%set_at(i, j, local_matrix(i,j))
        end do
      end do
    end subroutine

    subroutine test_without_vectors_${id}$(myid, matrix_a, rk, elpa, sv_min, sv_max, lower, upper)
      integer, intent(in) :: myid
      type(matrix_${id}$_t), intent(in) :: matrix_a
      integer, intent(in) :: rk, elpa
      real(${prec}$), intent(in) :: sv_min, sv_max
      real(${prec}$), intent(in), optional :: lower, upper

      integer :: info, i
      real(dp) :: accuracy
      real(${prec}$), allocatable :: singular_values(:), expected_singular_values(:)
      type(zolo_svd_t) :: svd

      allocate(singular_values(matrix_a%cols), expected_singular_values(matrix_a%cols))

      if (present(lower) .and. present(upper)) then
        call svd%set_bounds(lower, upper)
      end if
      call svd%set_descriptors(matrix_a%desc)
      call svd%set_option(ZSVD_R, rk)
      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 0)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 0)
      call svd%set_option(ZSVD_ELPA_SOLVER, elpa)
      call svd%set_option(ZSVD_DETECT_INVALID_VALUES, 1)

      info = svd%compute_svd(matrix_a%values, singular_values)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      do i = 1, matrix_a%cols
        expected_singular_values(i) = sv_min + ((sv_max - sv_min) / real(matrix_a%cols-1, kind=${prec}$)) * real(i-1, kind=${prec}$)
      end do
      if (myid == 0) then
        accuracy = maxval(abs(singular_values - expected_singular_values) / expected_singular_values)
        write(*,'(a,1x,g013.6e2)') "Accuracy of singular values:", accuracy
#:if prec == 'sp'
        if (accuracy > 1.e-4_dp) then
#:else
        if (accuracy > 1.e-12_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if
    end subroutine

    subroutine test_with_vectors_${id}$(myid, matrix_a, rk, elpa, lower, upper)
      integer, intent(in) :: myid
      type(matrix_${id}$_t), intent(in) :: matrix_a
      integer, intent(in) :: rk, elpa
      real(${prec}$), intent(in), optional :: lower, upper

      integer :: info
      real(dp) :: accuracy
      real(${prec}$), allocatable :: singular_values(:)
      type(zolo_svd_t) :: svd
      type(matrix_${id}$_t) :: matrix_left, matrix_right
      type(error_t) :: err

      allocate(singular_values(matrix_a%cols))

      call matrix_left%alloc_like(matrix_a, err)
      call matrix_right%alloc_like(matrix_a, err, matrix_a%cols, matrix_a%cols)
      if (err%has_error()) then
        call err%print()
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (present(lower) .and. present(upper)) then
        call svd%set_bounds(lower, upper)
      end if
      call svd%set_descriptors(matrix_a%desc, matrix_left%desc, matrix_right%desc)
      call svd%set_option(ZSVD_R, rk)
      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 1)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 1)
      call svd%set_option(ZSVD_COMPUTE_ACCURACY, 1)
      call svd%set_option(ZSVD_ELPA_SOLVER, elpa)
      call svd%set_option(ZSVD_DETECT_INVALID_VALUES, 1)

      info = svd%compute_svd(matrix_a%values, singular_values, matrix_left%values, matrix_right%values)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      accuracy = svd%get_accuracy()
      if (myid == 0) then
        write(*,'(a,1x,g013.6e2)') "Accuracy of SVD:", accuracy
#:if prec == 'sp'
        if (accuracy > 1.e-4_dp) then
#:else
        if (accuracy > 1.e-12_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if
    end subroutine

    subroutine test_only_left_vectors_${id}$(matrix_a, rk, elpa, lower, upper)
      type(matrix_${id}$_t), intent(in) :: matrix_a
      integer, intent(in) :: rk, elpa
      real(${prec}$), intent(in) :: lower, upper

      integer :: info
      real(${prec}$), allocatable :: singular_values(:)
      type(zolo_svd_t) :: svd
      type(matrix_${id}$_t) :: matrix_left, matrix_right
      type(error_t) :: err

      allocate(singular_values(matrix_a%cols))

      call matrix_left%alloc_like(matrix_a, err)
      call matrix_right%alloc_like(matrix_a, err, matrix_a%cols, matrix_a%cols)
      if (err%has_error()) then
        call err%print()
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call svd%set_bounds(lower, upper)
      call svd%set_descriptors(matrix_a%desc, matrix_left%desc, matrix_right%desc)
      call svd%set_option(ZSVD_R, rk)
      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 1)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 0)
      call svd%set_option(ZSVD_PRINT_ERRORS, 0)
      call svd%set_option(ZSVD_ELPA_SOLVER, elpa)
      call svd%set_option(ZSVD_DETECT_INVALID_VALUES, 1)

      info = svd%compute_svd(matrix_a%values, singular_values, matrix_left%values)
      if (info == 0) then
        write(*,'(a)') "No error in zolo_svd, although an error was expected"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call svd%set_option(ZSVD_PRINT_ERRORS, 1)

      info = svd%compute_svd(matrix_a%values, singular_values, matrix_left%values, matrix_right%values)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if
    end subroutine

    subroutine test_only_right_vectors_${id}$(matrix_a, rk, elpa, lower, upper)
      type(matrix_${id}$_t), intent(in) :: matrix_a
      integer, intent(in) :: rk, elpa
      real(${prec}$), intent(in) :: lower, upper

      integer :: info
      real(${prec}$), allocatable :: singular_values(:)
      type(zolo_svd_t) :: svd
      type(matrix_${id}$_t) :: matrix_right
      type(error_t) :: err

      allocate(singular_values(matrix_a%cols))

      call matrix_right%alloc_like(matrix_a, err, matrix_a%cols, matrix_a%cols)
      if (err%has_error()) then
        call err%print()
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call svd%set_bounds(lower, upper)
      call svd%set_descriptors(matrix_a%desc, desc_right_vectors=matrix_right%desc)
      call svd%set_option(ZSVD_R, rk)
      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 0)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 1)
      call svd%set_option(ZSVD_ELPA_SOLVER, elpa)
      call svd%set_option(ZSVD_DETECT_INVALID_VALUES, 1)

      info = svd%compute_svd(matrix_a%values, singular_values, right_singular_vectors=matrix_right%values)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if
    end subroutine

    subroutine test_invalid_${id}$(matrix_a)
      type(matrix_${id}$_t), intent(in) :: matrix_a

      integer :: info
      real(${prec}$), allocatable :: singular_values(:)
      type(zolo_svd_t) :: svd

      allocate(singular_values(matrix_a%cols))

      call svd%set_option(ZSVD_PRINT_ERRORS, 0)
      call svd%set_option(ZSVD_COMPUTE_LEFT_VECTORS, 0)
      call svd%set_option(ZSVD_COMPUTE_RIGHT_VECTORS, 0)

      info = svd%compute_svd(matrix_a%values, singular_values)
      if (info == 0) then
        write(*,'(a)') "Missing descriptors: No error, although an error was expected"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call svd%set_descriptors(matrix_a%desc)
      call svd%set_option(ZSVD_R, 0)

      info = svd%compute_svd(matrix_a%values, singular_values)
      if (info == 0) then
        write(*,'(a)') "Invalid r: No error, although an error was expected"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      call svd%set_option(ZSVD_R, 1)
      call svd%set_bounds(2._${prec}$, 1._${prec}$)

      info = svd%compute_svd(matrix_a%values, singular_values)
      if (info == 0) then
        write(*,'(a)') "Invalid lower and upper bounds: No error, although an error was expected"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if
    end subroutine

    subroutine test_solve_system_${id}$(matrix_a)
      type(matrix_${id}$_t), intent(in) :: matrix_a

      integer :: info, i, j
      ${dtype}$(${prec}$), allocatable :: vector_x(:), vector_x_ref(:), vector_b(:), vector_b_test(:), array_a(:,:)
      real(dp) :: accuracy
      type(zolo_svd_t) :: svd

      allocate(vector_x(matrix_a%cols), vector_x_ref(matrix_a%cols))
      allocate(vector_b(matrix_a%rows), vector_b_test(matrix_a%rows))

      do i = 1, matrix_a%cols
        vector_x_ref(i) = ${one}$ * i
      end do

      allocate(array_a(matrix_a%rows,matrix_a%cols))
      call matrix_a%copy_to_array(array_a)

      do i = 1, matrix_a%rows
        vector_b(i) = ${zero}$
        do j = 1, matrix_a%cols
          vector_b(i) = vector_b(i) + array_a(i,j) * vector_x_ref(j)
        end do
      end do

      call svd%set_descriptors(matrix_a%desc)
      info = svd%solve_system(matrix_a%values, vector_b, vector_x)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (myid == 0) then
        accuracy = maxval(abs(vector_x - vector_x_ref) / abs(vector_x_ref))
        write(*,'(a,1x,g013.6e2)') "Accuracy of solution (x):", accuracy
#:if prec == 'sp'
        if (accuracy > 5.e-3_dp) then
#:else
        if (accuracy > 1.e-11_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if

      if (matrix_a%rows == matrix_a%cols) then
        do i = 1, matrix_a%rows
          vector_b_test(i) = ${zero}$
          do j = 1, matrix_a%cols
            vector_b_test(i) = vector_b_test(i) + array_a(i,j) * vector_x(j)
          end do
        end do

        if (myid == 0) then
          accuracy = maxval(abs(vector_b - vector_b_test) / abs(vector_b_test))
          write(*,'(a,1x,g013.6e2)') "Accuracy of solution (b):", accuracy
#:if prec == 'sp'
          if (accuracy > 5.e-2_dp) then
#:else
          if (accuracy > 1.e-10_dp) then
#:endif
            call mpi_abort(MPI_COMM_WORLD, 1, info)
          end if
        end if
      end if
    end subroutine
#:endfor

end program
