#:include '../src/common.fypp'

module test_helper
  use constants, only: sp, dp

  contains

#:for prec in PRECS
    subroutine invert_sv_${prec}$(sv)
      real(${prec}$), intent(inout) :: sv(:)
      sv(:) = 1._${prec}$ / sv(:)
    end subroutine
#:endfor

end module

program test_zolo_svd_python_api
  use constants, only: sp, dp
#:if defined('WITH_MPI_MOD')
  use mpi
#:endif
  use test_helper
  use zolo_svd_python_api
  implicit none
#:if not defined('WITH_MPI_MOD')
  include 'mpif.h'
#:endif

  integer :: ierr, myid

  call mpi_init(ierr)
  call mpi_comm_rank(MPI_COMM_WORLD, myid, ierr)

#:for prec in PRECS
  call run_tests_${prec}$(myid, 17, 7)
  call run_tests_${prec}$(myid, 200, 200)
#:endfor

  if (myid == 0) then
    write(*,'(a)') "SUCCESS"
  end if

  call mpi_finalize(ierr)

  contains

#:for prec, id in zip(PRECS, ['s', 'd'])
    subroutine run_tests_${prec}$(myid, rows, cols)
      integer, intent(in) :: myid, rows, cols

      real(${prec}$), parameter :: sv_min = 17._${prec}$, sv_max = 27._${prec}$
      real(${prec}$), allocatable :: matrix(:,:), singular_values(:)

      allocate(matrix(rows,cols), singular_values(cols))

      call prepare_svd_${prec}$(sv_min, sv_max, matrix, singular_values)

      call test_without_vectors_${prec}$(myid, matrix, singular_values)

      call test_solve_system_${prec}$(myid, matrix)
    end subroutine

    !> Define singular values and construct the matrix using LAPACK's ?latms
    subroutine prepare_svd_${prec}$(sv_min, sv_max, matrix, singular_values)
      real(${prec}$), intent(in) :: sv_min, sv_max
      real(${prec}$), intent(inout) :: matrix(:,:), singular_values(:)

      integer :: rows, cols
      integer :: i, iseed(4), info
      real(${prec}$), allocatable :: work(:)

      rows = size(matrix, dim=1)
      cols = size(matrix, dim=2)

      allocate(work(4*cols))

      do i = 1, cols
        singular_values(i) = sv_min + ((sv_max - sv_min) / real(cols-1, kind=${prec}$)) * real(cols-i, kind=${prec}$)
      end do

      iseed(:) = 0
      call ${id}$latms(rows, cols, 'N', iseed, 'N', singular_values, 0, 0._${prec}$, 0._${prec}$, rows, cols, 'N', matrix, rows,&
          & work, info)
      if (info /= 0) then
        write(*,'(a)') "${id}$latms error"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if
    end subroutine

    subroutine test_without_vectors_${prec}$(myid, matrix, expected_singular_values)
      integer, intent(in) :: myid
      real(${prec}$), intent(in) :: matrix(:,:), expected_singular_values(:)

      integer :: cols, info
      real(dp) :: accuracy
      real(${prec}$), allocatable :: singular_values(:)

      cols = size(matrix, dim=2)

      allocate(singular_values(cols))

      info = compute_singular_values(matrix, singular_values, 1)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      singular_values(1:cols) = singular_values(cols:1:-1)

      if (myid == 0) then
        accuracy = maxval(abs(singular_values - expected_singular_values) / expected_singular_values)
        write(*,'(a,1x,g013.6e2)') "Accuracy of singular values:", accuracy
#:if prec == 'sp'
        if (accuracy > 1.e-4_dp) then
#:else
        if (accuracy > 1.e-12_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if
    end subroutine

    subroutine test_solve_system_${prec}$(myid, matrix)
      integer, intent(in) :: myid
      real(${prec}$), intent(in) :: matrix(:,:)

      integer :: rows, cols, i, j, info
      real(dp) :: accuracy
      real(${prec}$), allocatable :: vector_x(:), vector_x_ref(:), vector_b(:)

      rows = size(matrix, dim=1)
      cols = size(matrix, dim=2)

      allocate(vector_x(cols), vector_x_ref(cols))
      allocate(vector_b(rows))

      do i = 1, cols
        vector_x_ref(i) = real(i, kind=${prec}$)
      end do

      do i = 1, rows
        vector_b(i) = 0._${prec}$
        do j = 1, cols
          vector_b(i) = vector_b(i) + matrix(i,j) * vector_x_ref(j)
        end do
      end do

      info = solve_system(matrix, vector_b, vector_x, invert_sv_${prec}$, 1)
      if (info /= 0) then
        write(*,'(a)') "Error in zolo_svd"
        call mpi_abort(MPI_COMM_WORLD, 1, info)
      end if

      if (myid == 0) then
        accuracy = maxval(abs(vector_x - vector_x_ref) / abs(vector_x_ref))
        write(*,'(a,1x,g013.6e2)') "Accuracy of solution:", accuracy
#:if prec == 'sp'
        if (accuracy > 5.e-3_dp) then
#:else
        if (accuracy > 1.e-11_dp) then
#:endif
          call mpi_abort(MPI_COMM_WORLD, 1, info)
        end if
      end if
    end subroutine
#:endfor

end program
